/*
 The software in this package is distributed under the GNU General
 Public License version 2 (with a special exception described below).

 A copy of GNU General Public License (GPL) is included in this distribution,
 in the file COPYING.GPL.

 As a special exception, if other files instantiate templates or use macros
 or inline functions from this file, or you compile this file and link it
 with other works to produce a work based on this file, this file
 does not by itself cause the resulting work to be covered
 by the GNU General Public License.

 However the source code for this file must still be made available
 in accordance with section (3) of the GNU General Public License.

 This exception does not invalidate any other reasons why a work based
 on this file might be covered by the GNU General Public License.

 Christian Herdtweck, Intra2net AG 2015
 */

#include "dns/dnscache.h"

#include <fstream>
#include <logfunc.hpp>
#include <filefunc.hxx>   // I2n::file_exists
#include <boost/foreach.hpp>
#include <boost/bind.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/asio/placeholders.hpp>
#include <boost/serialization/serialization.hpp>
#include <boost/serialization/map.hpp>
#include <boost/serialization/string.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/archive/xml_oarchive.hpp>
#include <boost/archive/xml_iarchive.hpp>

#include "dns/dnsmaster.h"

using boost::bind;
using boost::posix_time::seconds;
using I2n::Logger::GlobalLogger;

namespace Config
{
    int SaveTimerSeconds = 60;
    int MaxRetrievalRecursions = 10;
}

DnsCache::DnsCache(const IoServiceItem &io_serv,
                   const std::string &cache_file)
    : IpCache()
    , CnameCache()
    , SaveTimer( *io_serv )
    , CacheFile( cache_file )
    , HasChanged( false )
{
    // load cache from file
    load_from_cachefile();

    // schedule next save
    (void) SaveTimer.expires_from_now( seconds( Config::SaveTimerSeconds ) );
    SaveTimer.async_wait( bind( &DnsCache::schedule_save, this,
                                boost::asio::placeholders::error ) );
}


DnsCache::~DnsCache()
{
    GlobalLogger.info() << "Dns Cache is being destructed";

    // save one last time without re-scheduling the next save
    save_to_cachefile();

    // cancel save timer
    SaveTimer.cancel();
}


void DnsCache::schedule_save(const boost::system::error_code &error)
{
    // just in case: ensure SaveTimer is cancelled
    SaveTimer.cancel();  // (will do nothing if already expired/cancelled)

    if ( error ==  boost::asio::error::operation_aborted )   // cancelled
    {
        GlobalLogger.error() << "DnsCache: SaveTimer was cancelled "
                             << "--> no save and no re-schedule of saving!";
        return;
    }
    else if (error)
    {
        GlobalLogger.error() << "DnsCache: Received error " << error
                             << " in schedule_save "
                             << "--> no save now but re-schedule saving";
    }
    else
        save_to_cachefile();

    // schedule next save
    (void) SaveTimer.expires_from_now( seconds( Config::SaveTimerSeconds ) );
    SaveTimer.async_wait( bind( &DnsCache::schedule_save, this,
                                boost::asio::placeholders::error ) );
}

void DnsCache::save_to_cachefile()
{
    if (!HasChanged)
    {
        GlobalLogger.info() << "DnsCache: skip saving because has not changed";
        return;
    }
    else if (CacheFile.empty())
    {
        GlobalLogger.warning()
                           << "DnsCache: skip saving because file name empty!";
        return;
    }

    try
    {
        std::ofstream ofs( CacheFile.c_str() );
        boost::archive::xml_oarchive oa(ofs);
        oa << boost::serialization::make_nvp("IpCache", IpCache);
        oa << boost::serialization::make_nvp("CnameCache", CnameCache);
        GlobalLogger.info() << "DnsCache: saved to cache file " << CacheFile;

        HasChanged = false;
    }
    catch (std::exception &exc)
    {
        GlobalLogger.warning() << "Saving failed: " << exc.what();
    }
}


void DnsCache::load_from_cachefile()
{
    if (CacheFile.empty())
    {
        GlobalLogger.warning()
                  << "DnsCache: cannot load because cache file name is empty!";
        return;
    }
    else if ( !I2n::file_exists(CacheFile) )
    {
        GlobalLogger.warning() << "DnsCache: cannot load because cache file "
                               << CacheFile << " does not exist!";
        return;
    }
    try
    {
        HostAddressVec cache;

        std::ifstream ifs( CacheFile.c_str() );
        boost::archive::xml_iarchive ia(ifs);

        ia >> boost::serialization::make_nvp("IpCache", cache);
        ia >> boost::serialization::make_nvp("CnameCache", cache);
        GlobalLogger.info() << "DnsCache: loaded from file " << CacheFile;
    }
    catch (boost::archive::archive_exception &exc)
    {
        GlobalLogger.warning() << "DnsCache: archive exception loading from "
                               << CacheFile << ": " << exc.what();
    }
    catch (std::exception &exc)
    {
        GlobalLogger.warning() << "DnsCache: exception while loading from "
                               << CacheFile << ": " << exc.what();
    }
}

void DnsCache::update(const std::string &hostname,
                      const HostAddressVec &new_data)
{
    GlobalLogger.info() << "DnsCache: update IPs for " << hostname
                        << " to " << new_data.size() << "-list";
    IpCache[hostname] = new_data;
    HasChanged = true;
}


void DnsCache::update(const std::string &hostname,
                      const Cname &cname)
{
    GlobalLogger.info() << "DnsCache: update CNAME for " << hostname
                        << " to " << cname.first;
    CnameCache[hostname] = cname;
    HasChanged = true;
}


void DnsCache::update(const std::string &hostname,
                          const uint32_t new_ttl)
{
    GlobalLogger.info() << "DnsCache: ensure TTL for IPs for " << hostname
                        << " is below " << new_ttl << "s";
    HostAddressVec ips = IpCache[hostname];
    TimeToLive current_ttl;
    BOOST_FOREACH( HostAddress &addr, ips )
    {
        current_ttl = addr.get_ttl();
        if (current_ttl.get_value() > new_ttl)
        {
            current_ttl.set_value(new_ttl);
            addr.set_ttl(current_ttl);
            HasChanged = true;
        }
    }
    IpCache[hostname] = ips;
}


HostAddressVec DnsCache::get_ips(const std::string &hostname,
                                 const bool check_up_to_date)
{
    HostAddressVec result = IpCache[hostname];
    if (check_up_to_date)
    {
        HostAddressVec result_up_to_date;
        int threshold = DnsMaster::get_instance()
                        ->get_resolved_ip_ttl_threshold();
        BOOST_FOREACH( const HostAddress &addr, result )
        {
            if (addr.get_ttl().get_updated_value() > threshold)
                result_up_to_date.push_back(addr);
        }
        GlobalLogger.debug() << "DnsCache: From cached list of size "
             << result.size() << " return " << result_up_to_date.size()
             << " since rest out of date";
        result = result_up_to_date;
    }
    GlobalLogger.info() << "DnsCache: request IPs for " << hostname
                        << " --> " << result.size() << "-list";
    return result;
}

std::string DnsCache::get_cname(const std::string &hostname,
                                 const bool check_up_to_date)
{
    Cname result_obj = CnameCache[hostname];
    GlobalLogger.info() << "DnsCache: request CNAME for " << hostname
                        << " --> \"" << result_obj.first << "\"";
    if (check_up_to_date)
    {
        if (result_obj.second.get_updated_value() > DnsMaster::get_instance()
                                              ->get_resolved_ip_ttl_threshold())
            return result_obj.first;
        else
        {
            GlobalLogger.debug() << "DnsCache: Cname is out of date";
            return "";
        }
    }
    else
        return result_obj.first;
}

// underlying assumption in this function: for a hostname, the cache has either
// a list of IPs saved or a cname saved, but never both
HostAddressVec DnsCache::get_ips_recursive(const std::string &hostname,
                                           const bool check_up_to_date)
{
    std::string current_host = hostname;
    HostAddressVec result = get_ips(current_host);
    int n_recursions = 0;
    while ( result.empty() )
    {
        current_host = get_cname(current_host, check_up_to_date);
        if (current_host.empty())
            break;
        else if (++n_recursions >= Config::MaxRetrievalRecursions)
        {
            GlobalLogger.warning() << "DnsCache: reached recursion limit of "
                << n_recursions << " in recursive IP retrieval!";
            break;
        }
        else
            result = get_ips(current_host, check_up_to_date);
    }
    return result;
}

// (created using vim -- the world's best text editor)

